#!/usr/bin/env python
"""SPF (Sender-Permitted From) test script.

Copyright (c) 2003, Terence Way
This module is free software, and you may redistribute it and/or modify
it under the same terms as Python itself, so long as this copyright message
and disclaimer are retained in their original form.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF
THIS CODE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.  THE CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
AND THERE IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
"""

# Changes:
#    9-dec-2003, v1.1, Meng Weng Wong added PTR code, THANK YOU
#   11-dec-2003, v1.2, ttw added macro expansion, exp=, and redirect=

__author__ = "Terence Way"
__email__ = "terry@wayforward.net"
__version__ = "1.2: December 11, 2003"
MODULE = 'spf'

USAGE = """To check an incoming mail request:
    % python spf.py {ip} {sender} {helo}
    % python spf.py 69.55.226.139 tway@optsw.com mx1.wayforward.net

To test an SPF record:
    % python spf.py "v=spf1..." {ip} {sender-email}
    % python spf.py "v=spf1 +mx +ip 10.0.0.1 -all" 10.0.0.1 tway@foo.com a    

To test this script:
    % python spf.py
"""

import re
import time
import DNS	# http://pydns.sourceforge.net

# 32-bit IPv4 address mask
MASK = 0xFFFFFFFFL

# Regular expression to look for modifiers
RE_MODIFIER = re.compile(r'^([a-zA-Z]+)=')

# Regular expression to find macro expansions
RE_CHAR = re.compile(r'%(%|_|-|(\{[a-zA-Z][0-9]*r?[^\}]*\}))')

# Regular expression to break up a macro expansion
RE_ARGS = re.compile(r'([0-9]*)(r?)([^0-9a-zA-Z]*)')

# Local parts and senders have their delimiters replaced with '.' during
# macro expansion
#
JOINERS = {'l': '.', 's': '.'}

def check(i, s, h):
	"""Test an incoming MAIL FROM:<s>, from a client with ip address i.
	"""
	try:
		q = query(i=i, s=s, h=h)
		return q.check(get_spf(q.d))
	except DNS.DNSError:
		return ('error', 250, 'SPF DNS Error')

class query(object):
	"""A query object keeps the relevant information about a single SPF
	query:

	i: ip address of SMTP client
	s: sender declared in MAIL FROM:<>
	l: local part of sender s
	d: domain part of sender s
	h: EHLO/HELO domain
	v: 'in-addr' for IPv4 clients and 'ip6' for IPv6 clients
	t: current timestamp
	p: SMTP client domain name

	This is also, by design, the same variables used in SPF macro
	expansion.
	"""
	def __init__(self, i, s, h):
		self.i, self.s, self.h = i, s, h
		self.l, self.d = split_email(s, h)
		self.t = str(int(time.time()))
		self.v = 'in-addr'
		self.p = None

	def getp(self):
		if not self.p:
			p = get_ptr(self.i)
			if len(p) > 0:
				self.p = p[0]
			else:
				self.i = self.i
		return self.p

	def check(self, spf):
		return self.check1(spf, self.d, 0)

	def check1(self, spf, domain, recursion):
		# spf rfc: 3.7 Processing Limits
		#
		if recursion > 10:
			return ('unknown', 250, 'SPF recursion limit exceeded')
		try:
			tmp, self.d = self.d, domain
			return self.check0(spf, recursion)
		finally:
			self.d = tmp

	def check0(self, spf, recursion):
		"""Test this query information against SPF text."""

		if not spf:
			return ('unknown', 200, 'no SPF record')

		# split string by whitespace, drop the 'v=spf1'
		#
		spf = spf.split()[1:]

		exp = 'access denied'
		unknown_exp = 'SPF unknown'

		redirect = None

		# Look for modifiers
		#
		for m in spf:
			m = RE_MODIFIER.split(m)[1:]
			if len(m) != 2: continue

			if m[0] == 'exp':
				exp = self.expand(m[1])
				exp = self.expand(''.join(get_txt(exp)))
				unknown_exp = exp
			elif m[0] == 'redirect':
				redirect = self.expand(m[1])

			# spf rfc: 3.6 Unrecognized Mechanisms and Modifiers

		# Look for mechanisms
		#
		for m in spf:
			if RE_MODIFIER.match(m): continue

			m, arg, cidrlength = parse_mechanism(m, self.d)

			if m in ['a', 'mx', 'ptr', 'exists', 'include']:
				arg = self.expand(arg)

			if m == 'include':
				if arg != self.d:
					result = self.check1(get_spf(arg), arg,
					                     recursion  + 1)
					if result[0] == 'pass' or \
					   result[0] != 'fail':
						return result
				continue

			if m[0] == '-':
				result = ('deny', 550, exp)
				m = m[1:]
			elif m[0] == '?':
				result = ('unknown', 250, unknown_exp)
				m = m[1:]
			else:
				result = ('pass', 250, 'sender SPF validated')
				if m[0] == '+':
					m = m[1:]

			if m == 'all':
				return result

			elif m == 'exists':
				if len(get_a(arg)) > 0:
					return result

			elif m == 'a':
				if cidrmatch(self.i, get_a(arg), cidrlength):
					return result

			elif m == 'mx':
				if cidrmatch(self.i, get_mx(arg), cidrlength):
					return result

			elif m == 'ip4' and arg != self.d:
				if cidrmatch(self.i, [arg], cidrlength):
					return result

			elif m == 'ptr':
				if domainmatch(validated_ptrdnames(self.i),
				               arg):
					return result

			else:
				# unknown mechanisms cause immediate unknown
				# abort results
				return ('unknown', 250, unknown_exp)

		if redirect:
			return self.check1(get_spf(redirect), redirect,
			                   recursion + 1)

		# no mechanisms at all cause unknown result
		return ('unknown', 250, unknown_exp)

	def expand(self, str):
		"""Do SPF RFC macro expansion.

		Examples:
		>>> q = query(s='strong-bad@email.example.com',
		...           h='mx.example.org', i='192.0.2.3')
		>>> q.p = 'mx.example.org'

		>>> q.expand('%{d}')
		'email.example.com'

		>>> q.expand('%{d4}')
		'email.example.com'

		>>> q.expand('%{d3}')
		'email.example.com'

		>>> q.expand('%{d2}')
		'example.com'

		>>> q.expand('%{d1}')
		'com'

		>>> q.expand('%{p}')
		'mx.example.org'

		>>> q.expand('%{p2}')
		'example.org'

		>>> q.expand('%{dr}')
		'com.example.email'
	
		>>> q.expand('%{d2r}')
		'example.email'

		>>> q.expand('%{l}')
		'strong-bad'

		>>> q.expand('%{l-}')
		'strong.bad'

		>>> q.expand('%{lr}')
		'strong-bad'

		>>> q.expand('%{lr-}')
		'bad.strong'

		>>> q.expand('%{l1r-}')
		'strong'

		>>> q.expand('%{ir}.%{v}._spf.%{d2}')
		'3.2.0.192.in-addr._spf.example.com'

		>>> q.expand('%{lr-}.lp._spf.%{d2}')
		'bad.strong.lp._spf.example.com'

		>>> q.expand('%{lr-}.lp.%{ir}.%{v}._spf.%{d2}')
		'bad.strong.lp.3.2.0.192.in-addr._spf.example.com'

		>>> q.expand('%{ir}.%{v}.%{l1r-}.lp._spf.%{d2}')
		'3.2.0.192.in-addr.strong.lp._spf.example.com'

		>>> q.expand('%{p2}.trusted-domains.example.net')
		'example.org.trusted-domains.example.net'

		>>> q.expand('%{p2}.trusted-domains.example.net')
		'example.org.trusted-domains.example.net'

		"""
		end = 0
		result = ''
		for i in RE_CHAR.finditer(str):
			result += str[end:i.start()]
			macro = str[i.start():i.end()]
			if macro == '%%':
				result += '%'
			elif macro == '%_':
				result += ' '
			elif macro == '%-':
				result += '%20'
			else:
				letter = macro[2].lower()
				if letter == 'p':
					self.getp()
				expansion = getattr(self, letter, '')
				if expansion:
					result += expand_one(expansion,
						macro[3:-1],
					        JOINERS.get(letter))

			end = i.end()
		return result + str[end:]

def split_email(s, h):
	"""Given a sender email s and a HELO domain h, create a valid tuple
	(l, d) local-part and domain-part.

	Examples:
	>>> split_email('', 'wayforward.net')
	('postmaster', 'wayforward.net')

	>>> split_email('foo.com', 'wayforward.net')
	('postmaster', 'foo.com')

	>>> split_email('terry@wayforward.net', 'optsw.com')
	('terry', 'wayforward.net')
	"""
	if not s:
		return 'postmaster', h
	else:
		parts = s.split('@', 1)
		if len(parts) == 2:
			return tuple(parts)
		else:
			return 'postmaster', s

def parse_mechanism(str, d):
	"""Breaks A, MX, IP4, and PTR mechanisms into a (name, domain,
	cidr) tuple.  The domain portion defaults to d if not present,
	the cidr defaults to 32 if not present.

	Examples:
	>>> parse_mechanism('a', 'foo.com')
	('a', 'foo.com', 32)

	>>> parse_mechanism('a:bar.com', 'foo.com')
	('a', 'bar.com', 32)

	>>> parse_mechanism('a/24', 'foo.com')
	('a', 'foo.com', 24)

	>>> parse_mechanism('a:bar.com/16', 'foo.com')
	('a', 'bar.com', 16)
	"""
	a = str.split('/')
	if len(a) == 2:
		a, port = a[0], int(a[1])
	else:
		a, port = str, 32

	b = a.split(':')
	if len(b) == 2:
		return b[0], b[1], port
	else:
		return a, d, port

def get_spf(domainname):
	"""Get the SPF record recorded in DNS for a specific domain name.
	Returns None if not found, or if more than one record is found.
	"""
	a = [txt for txt in get_txt(domainname) if txt.startswith('v=spf1')]

	if len(a) == 1:
		return a[0]
	else:
		return None

def get_txt(domainname):
	return [t for a in get_dns(domainname, 'txt') for t in a]

def get_mx(domainname):
	"""Get a list of IP addresses for all MX exchanges for a domain name.
	"""
	return [a for mx in get_dns(domainname, 'mx') for a in get_a(mx[1])]

def get_a(domainname):
	"""Get a list of IP addresses for a domainname."""
	return get_dns(domainname, 'a')

def get_ptr(i):
	"""Get a list of domain names for an IP address."""
	return get_dns(reverse_dots(i) + ".in-addr.arpa", 'ptr')

def get_dns(domainname, qtype):
	req = DNS.DnsRequest(domainname, qtype=qtype)
	resp = req.req()
	return [a['data'] for a in resp.answers]

def reverse_dots(name):
	"""Reverse dotted IP addresses or domain names.

	Example:
	>>> reverse_dots('192.168.0.145')
	'145.0.168.192'

	>>> reverse_dots('email.example.com')
	'com.example.email'
	"""
	a = name.split('.')
	a.reverse()
	return '.'.join(a)

def domainmatch(ptrdnames, domainsuffix):
	"""grep for a given domain suffix against a list of validated PTR
	domain names.

	Examples:
	>>> domainmatch(['FOO.COM'], 'foo.com')
	1

	>>> domainmatch(['moo.foo.com'], 'FOO.COM')
	1

	>>> domainmatch(['moo.bar.com'], 'foo.com')
	0

	"""
	domainsuffix = domainsuffix.lower()
	for ptrdname in ptrdnames:
		ptrdname = ptrdname.lower()

		if ptrdname == domainsuffix or \
		   ptrdname.endswith('.' + domainsuffix):
			return True

	return False

def validated_ptrdnames(i):
	"""Figure out the validated PTR domain names for a given IP address.
	"""
	return [p for p in get_ptr(i) if i in get_a(p)]

def cidrmatch(i, ipaddrs, cidr_length = 32):
	"""Match an IP address against a list of other IP addresses.

	Examples:
	>>> cidrmatch('192.168.0.45', ['192.168.0.44', '192.168.0.45'])
	1

	>>> cidrmatch('192.168.0.43', ['192.168.0.44', '192.168.0.45'])
	0

	>>> cidrmatch('192.168.0.43', ['192.168.0.44', '192.168.0.45'], 24)
	1
	"""
	c = cidr(i, cidr_length)
	for ip in ipaddrs:
		if cidr(ip, cidr_length) == c:
			return True
	return False

def cidr(i, n):
	"""Convert an IP address string with a CIDR mask into a 32-bit
	integer.

	Examples:
	>>> DNS.bin2addr(cidr('192.168.5.45', 32))
	'192.168.5.45'
	>>> DNS.bin2addr(cidr('192.168.5.45', 24))
	'192.168.5.0'
	>>> DNS.bin2addr(cidr('192.168.0.45', 8))
	'192.0.0.0'
	"""
	return ~(MASK >> n) & MASK & DNS.addr2bin(i)

def expand_one(expansion, str, joiner):
	if not str:
		return expansion
	len, reverse, delimiters = RE_ARGS.split(str)[1:4]
	if not delimiters:
		delimiters = '.'
	expansion = split(expansion, delimiters, joiner)
	if reverse: expansion.reverse()
	if len: expansion = expansion[-int(len)*2+1:]
	return ''.join(expansion)

def split(str, delimiters, joiner=None):
	"""Split a string into pieces by a set of delimiter characters.  The
	resulting list is delimited by joiner, or the original delimiter if
	joiner is not specified.

	Examples:
	>>> split('192.168.0.45', '.')
	['192', '.', '168', '.', '0', '.', '45']

	>>> split('terry@wayforward.net', '@.')
	['terry', '@', 'wayforward', '.', 'net']

	>>> split('terry@wayforward.net', '@.', '.')
	['terry', '.', 'wayforward', '.', 'net']
	"""
	result, element = [], ''
	for c in str:
		if c in delimiters:
			result.append(element)
			element = ''
			if joiner:
				result.append(joiner)
			else:
				result.append(c)
		else:
			element += c
	result.append(element)
	return result

def _test():
	import doctest, spf
	return doctest.testmod(spf)

if __name__ == '__main__':
	import sys
	if len(sys.argv) == 4:
		DNS.DiscoverNameServers()
		print check(i=sys.argv[1], s=sys.argv[2], h=sys.argv[3])
	elif len(sys.argv) == 5:
		DNS.DiscoverNameServers()
		i, s, h = sys.argv[2:]
		q = query(i=i, s=s, h=h)
		q.check(sys.argv[1])
	elif len(sys.argv) == 1:
		_test()
	else:
		print USAGE
